// sigapass3 - 	2nd order all-pass style (like wahwah) ---   
// 				zack settel 11/93  
//				msp version 1998 dudas
//				modified 2p2z structure integrated into perform method

#include "ext.h"
#include "z_dsp.h"
#include <math.h>

//#define RES_ID 8896
//#define TWOPI 6.28318530717959 // already defined in MSP headers

typedef struct _sigapass3
{
	t_pxobject x_obj;
	double c_state1;	/* previous output 				*/
	double c_state2;	/* previous previous output 	*/
	double c_fconv;		/* sample rate conv. factor*/
	int c_vsize;		/* vector size */
} t_sigapass3;

void *sigapass3_class;

t_int *sigapass3_perform(t_int *w);
void sigapass3_dsp(t_sigapass3 *x, t_signal **sp, short *count);
void sigapass3_clear(t_sigapass3 *x);
void sigapass3_assist(t_sigapass3 *x, void *b, long m, long a, char *s);
void *sigapass3_new(long n);

void main(void)
{
	setup((t_messlist **)&sigapass3_class, (method)sigapass3_new, (method)dsp_free, (short)sizeof(t_sigapass3), 0L, A_DEFLONG, 0);
	addmess((method)sigapass3_dsp, "dsp", A_CANT, 0);
	addmess((method)sigapass3_clear, "clear", 0L);
	//addfloat((method)sigapass3_float); 		/* paramSig: base freq. */
	//addftx((method)sigapass3_ft1, 1); 		/* paramSig  notch-width */
	
	addmess((method)sigapass3_assist, "assist", A_CANT, 0);	
	//rescopy('STR#',RES_ID);
	
	dsp_initclass();
}

void sigapass3_clear(t_sigapass3 *x)
{
	x->c_state1 = 0.;	/* initialize 2p2z filter state */	 
	x->c_state2 = 0.;
}

/* radian frequency clips to from 0 to 2/3.14159  */
/* updates coefs twice per vector */

t_int *sigapass3_perform(t_int *w)
{
	t_float *in = (t_float *)(w[1]);
	t_float *out = (t_float *)(w[2]);
	t_sigapass3 *x = (t_sigapass3 *)(w[3]);
	t_float *cfreq = (t_float *)(w[4]);
	t_float *qval = (t_float *)(w[5]);
	int vs = (int)(w[6]);
	int n;
		
	double cos_theta, tmp;
	double theta;
	double Q;//, notchwidth;

  	double ym1, ym2, ym0;
  	double a, b;
  	double ff0, ff1;

	if (x->x_obj.z_disabled)
		goto out;

	ym1 = x->c_state1;	// get previous state from object instance
	ym2 = x->c_state2;
	
	/* calculate values for first half of vector */  
	n = vs * .5;
  	theta = x->c_fconv * *cfreq;  
	Q =  (*qval <= .00001) ? .00001: *qval;
	if (theta > 2.) theta = 2.;
	else if (theta < .0001) theta = .0001; /* safety */
	cos_theta = 1. - .5 * theta * theta;
 	
	/* set up filter */
	tmp = exp((-theta *.5)/Q);
	a = 2*cos_theta * tmp;	/* find A for C1 */
	ff1 = -a;  				/* -A for D1 	 */
	ff0 = tmp*tmp;  		/* find B for D0 */
	b = -ff0;				/* -B for C2 	 */
	
	++n; // increment n once for pre-decrement
	
 	/* whack first half of vector with the 2p2z~ calculation */
	while (--n)
	{
		ym0 = a * ym1 + b * ym2 + *++in + 1E-20;
		*++out = ff0 * ym0 + ff1 * ym1 + ym2;
		ym2 = ym1;
		ym1 = ym0;
    } 
	
	/* calculate new values for second half of vector */
	n = vs * .5;
	theta = x->c_fconv * *(cfreq+n);
	Q =  (*(qval+n) <= .00001) ? .00001: *(qval+n);
	if (theta > 2.) theta = 2.;
	else if (theta < .0001) theta = .0001; /* safety */
	cos_theta = 1. - .5 * theta * theta;
  
	/* set up filter again */
	tmp = exp((- theta *.5)/Q);
	a = 2*cos_theta * tmp;	/* find A for C1 */
	ff1 = -a;  				/* -A for D1 	 */
	ff0 = tmp*tmp;  		/* find B for D0 */
	b = -ff0;				/* -B for C2 	 */
 
 	++n; // increment n once for pre-decrement
 	
	/* whack second half with the 2p2z~ calculation */
	while (--n)
	{
		ym0 = a * ym1 + b * ym2 + *++in + 1E-20;
		*++out = ff0 * ym0 + ff1 * ym1 + ym2;
		ym2 = ym1;
		ym1 = ym0;
    }
    
	x->c_state1 = ym1; // save current state to object instance
	x->c_state2 = ym2; 

out:
	return (w+7);
}

//-----------------------------------------------------------------------

void sigapass3_dsp(t_sigapass3 *x, t_signal **sp, short *count)
{
	x->c_fconv = TWOPI / sp[0]->s_sr;	// 2pi/SR
	x->c_vsize = sp[0]->s_n; 			// VS
	
	x->c_state1 = 0.;	/* initialize 2p2z filter state */	 
	x->c_state2 = 0.;
	
	dsp_add(sigapass3_perform, 6, sp[0]->s_vec-1, sp[3]->s_vec-1, x, sp[1]->s_vec, sp[2]->s_vec, sp[0]->s_n);

}

//-----------------------------------------------------------------------

void sigapass3_assist(t_sigapass3 *x, void *b, long m, long a, char *s)
{
	//assist_string(RES_ID,m,a,1,4,s);
	if (m==1) {
		switch (a) {
			case 0: sprintf(s,"(Signal) Input Signal"); break;
			case 1: sprintf(s,"(Signal) Turnover Frequency"); break;
			case 2: sprintf(s,"(Signal) Transition Width (Q)"); break;
		}
	} 
	else if (m==2) {
		switch (a) {
			case 0: sprintf(s,"(Signal) Phase Modified Signal Out"); break;
		}
	}
}

void *sigapass3_new(long n)
{
	t_sigapass3 *x = (t_sigapass3 *)newobject(sigapass3_class);
	dsp_setup((t_pxobject *)x,3);
    outlet_new((t_pxobject *)x, "signal");
    
	x->c_fconv = 44100; /* default SR */

	x->c_state1 = 0.;	/* initialize 2p2z filter state */	 
	x->c_state2 = 0.;
	
	return (x);
}


