// sigarctan2 - arctangent2 function   
// 				zack settel 11/93  
//				msp version 1998 dudas

#include "ext.h"
#include "z_dsp.h"
#include <math.h>

//#define RES_ID 6668
#define TABLE_SIZE 4096
#define PI_OVER_TWO 1.5707963
//#define PI 3.14159265 // defined in MSP headers
#define NPI -3.14159265
#define NPI_OVER_TWO -1.5707963
#define VERSION "arctan2~  1-94 bilp & zeep"

typedef struct _sigarctan2
{
	t_pxobject x_obj;
	int a_memory;
	double *a_table;
} t_sigarctan2;

void *sigarctan2_class;

t_int *sigarctan2_perform(t_int *w);
void sigarctan2_dsp(t_sigarctan2 *x, t_signal **sp, short *count);
void sigarctan2_assist(t_sigarctan2 *x, void *b, long m, long a, char *s);
void sigarctan2_free(t_sigarctan2 *x);
void *sigarctan2_new(long n);

void main(void)
{
	setup((t_messlist **)&sigarctan2_class, (method)sigarctan2_new, (method)sigarctan2_free, (short)sizeof(t_sigarctan2), 0L, A_DEFLONG, 0);
	addmess((method)sigarctan2_dsp, "dsp", A_CANT, 0);
	
	addmess((method)sigarctan2_assist, "assist", A_CANT, 0);	
	//rescopy('STR#',RES_ID);
	
	post("%s", VERSION);
	dsp_initclass();
}


t_int *sigarctan2_perform(t_int *w)
{
	t_float *yval = (t_float *)(w[1]);
	t_float *xval = (t_float *)(w[2]);
	t_float *out = (t_float *)(w[3]);
	t_sigarctan2 *x = (t_sigarctan2 *)(w[4]);
	int n = (int)(w[5]);
	
	double  inval, outval, offset, where;
	double ax,ay;
	int bin;
	double *table;
	
	if (x->x_obj.z_disabled)
		goto out;
		
	table = x->a_table;
	while (--n) 
	{
		ay = (*yval < 0) ? 0-*yval:*yval;
		ax = (*xval < 0) ? 0-*xval:*xval;
		
		if(ay<ax)	/* normal atan lookup - atan(y/x) */
		{
			inval = ay / ax;
			where = (TABLE_SIZE * inval);
			bin = (int) (where);
			
			offset = where - bin;
			outval = (1-offset)*(*(table +  bin))+offset*(*(table+bin+1));
		}
		else		/* pi/2 - atan(x/y) */
		{
			inval = (ay==0) ? 0: ax / ay; /* force phase angle of 0 for case: x=y=0 */
			where = (TABLE_SIZE * inval);
			bin = (int) (where);
			
			offset = where - bin;
			outval = PI_OVER_TWO - ((1-offset)*(*(table + bin))+offset*(*(table + bin+1)));
		}

		if (*xval<0.)
		{
		    if (*yval >=0.)	/* -,+ 2nd */
		    {
				*out = PI - outval;
		    }			
		    else		/* -,-  3rd */
		    {			
				*out = NPI + outval;
		    }
		}
		else
		{
		    if (*yval < 0.)	/* +,-  4th */
		    {			
				*out = 0 - outval;
		    }
		    else		/* +,-  1st */
		    {			
				*out = outval;
		    }
		}
		*++yval;
		*++xval;
		*++out; 
	} 
out:
	return (w+6);	
}

//-----------------------------------------------------------------------

void sigarctan2_dsp(t_sigarctan2 *x, t_signal **sp, short *count)
{
	dsp_add(sigarctan2_perform, 5, sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec, x, sp[0]->s_n+1);
	// no need to pre-decrement s_vecs because ++vec at end of while
}

//-----------------------------------------------------------------------

void sigarctan2_assist(t_sigarctan2 *x, void *b, long m, long a, char *s)
{
	//assist_string(RES_ID,m,a,1,3,s);
	if (m==1) {
		switch (a) {
			case 0: sprintf(s,"(Signal) Y value"); break;
			case 1: sprintf(s,"(Signal) X value"); break;
		}
	} 
	else if (m==2) {
		switch (a) {
			case 0: sprintf(s,"(Signal) arctan2(Y,X)"); break;
		}
	}
}

//-----------------------------------------------------------------------

void sigarctan2_free(t_sigarctan2 *x)
{
	t_freebytes(x->a_table,(TABLE_SIZE+2) * sizeof(double));
	dsp_free((t_pxobject *)x);
}

//-----------------------------------------------------------------------

void *sigarctan2_new(long n)
{
	double *table;
	int i;
	
	t_sigarctan2 *x = (t_sigarctan2 *)newobject(sigarctan2_class);
	dsp_setup((t_pxobject *)x,2);
    outlet_new((t_pxobject *)x, "signal");
    
	
	x->a_table = t_getbytes((TABLE_SIZE+2)*sizeof(double));
	table = x->a_table;
	for(i=0;i<=TABLE_SIZE;i++)
	{
		*table++ = atan(((double)i)/((double)TABLE_SIZE));
	}
	x->a_table[TABLE_SIZE+1] = atan(1.);	/* extra point for interp. */
	
	return (x);
}


