#include "ext.h"
#include "z_dsp.h"
#include "math.h"

#define DELAY_ALIGN	16		/* alignment needed, in bytes, for delay line */

typedef struct delbuf
{
	float *b_shit;
	float b_conv;
	long b_phase;
	long b_nsamps;
} t_delbuf;

typedef struct _delwrite
{
	t_pxobject x_obj;
	t_symbol *x_sym;
	struct _delwrite *x_next;	/* list of delwrites hanging from symbol */
	t_delbuf x_buf;
	char *x_malloc;
	long x_on;				// hack, has it been put on chain yet?
} t_delwrite;

typedef struct _delread
{
	t_pxobject x_obj;
	t_symbol *x_sym;
	float x_rawdeltime;		/* nominal delay time */
	float x_delonset;		/* 0 if read before write, vecsize/SR otherwise */
	float x_deltime;		/* difference of two above */
	t_delbuf *x_buf;
} t_delread;

void cu_delread(float *out, t_delbuf *b, long n, double deltime);
t_int *delread_perform(t_int *w);
void delread_float(t_delread *x, double f);
void delread_int(t_delread *x, long n);
void delread_dsp(t_delread *x, t_signal **sp, short *count);
void delread_assist(t_delread *x, void *b, long m, long a, char *s);
void *delread_new(t_symbol *s, double f, long hack);

void *delread_class;
t_symbol *ps_delwrite_class;

void main(void)
{
	setup((t_messlist **)&delread_class, (method)delread_new, (method)dsp_free, (short)sizeof(t_delread), 
		0L, A_SYM, A_DEFFLOAT, A_DEFLONG, 0);
	addmess((method)delread_dsp, "dsp", A_CANT, 0);
	addmess((method)delread_assist, "assist", A_CANT, 0);
	addfloat((method)delread_float);
	addint((method)delread_int);
	dsp_initclass();
	ps_delwrite_class = gensym("delwrite_class");
}

void cu_delread(float *out, t_delbuf *b, long n, double deltime)
{
	long offset = deltime * b->b_conv - n;
	float *bp;
	
	if (offset > b->b_nsamps)
		offset = b->b_nsamps;
	else if (offset < n)
		offset = n;
	bp = b->b_shit + b->b_phase - offset;
	if (bp < b->b_shit) {
		bp += b->b_nsamps;
		//post("shit at end %.5f,%.5f,%.5f,%.5f",*bp,*(bp+1),*(bp+2),*(bp+3));
	}
	sysmem_copyptr((char *)bp, (char *)out, n * sizeof(float));
}

t_int *delread_perform(t_int *w)
{
	float *out;
	int n;
	t_delread *x = (t_delread *)(w[1]);
	out = (float *)(w[2]);
	n = w[3];
	
	if (x->x_obj.z_disabled)
		goto out;
	cu_delread(out, x->x_buf, n, x->x_deltime);
out:
	return w + 4;
}
	
void delread_float(t_delread *x, double f)
{
	x->x_deltime = (x->x_rawdeltime = f) - x->x_delonset;
}

void delread_int(t_delread *x, long n)
{
	delread_float(x, (double)n);
}

void delread_dsp(t_delread *x, t_signal **sp, short *count)
{
	t_delwrite *wt;
	
	wt = (t_delwrite *)(x->x_sym->s_thing);
	if (wt && ob_messlist(wt) == *(t_messlist **)ps_delwrite_class->s_thing) {
		x->x_delonset = wt->x_on ? 0 : sp[0]->s_n / (sp[0]->s_sr * .001);
		x->x_deltime = x->x_rawdeltime - x->x_delonset;
		x->x_buf = &wt->x_buf;
		dsp_add(delread_perform, 3, x, sp[1]->s_vec, sp[1]->s_n);
	} else
		error("delread~ %s: no corresponding delwrite~",x->x_sym->s_name);
}

void delread_assist(t_delread *x, void *b, long m, long a, char *s)
{
	if (m==1) {
		switch (a) {
			case 0: sprintf(s,"(dummy) old order forcing inlet"); break;
		}
	} 
	else if (m==2) {
		switch (a) {
			case 0: sprintf(s,"(signal) delayed signal out"); break;
		}
	}
}

void *delread_new(t_symbol *s, double f, long hack)
{
	t_delread *x = (t_delread *)newobject(delread_class);
	dsp_setup((t_pxobject *)x,1);			// order forcing inlet
	x->x_sym = s;
	x->x_rawdeltime = x->x_deltime = f;
	x->x_delonset = 0.;
	outlet_new((t_object *)x, "signal");
	return (x);
}
