#include "ext.h"
#include "z_dsp.h"
#include <math.h>

#define DELAY_ALIGN	16		/* alignment needed, in bytes, for delay line */
#define SAMP_NBITS 8
#define SAMP_NPOINTS (1 << SAMP_NBITS)

#define SYS_MINBLKSIZE 2 // no longer in z_dsp.h

typedef struct sampfilt
{
  float f1, f2, f3, f4;
} t_sampfilt;

typedef struct delbuf
{
	float *b_shit;
	float b_conv;
	long b_phase;
	long b_nsamps;
} t_delbuf;

typedef struct _delwrite
{
	t_pxobject x_obj;
	t_symbol *x_sym;
	struct _delwrite *x_next;	/* list of delwrites hanging from symbol */
	t_delbuf x_buf;
	char *x_malloc;
	long x_on;				// hack, has it been put on chain yet?
} t_delwrite;

typedef struct _vd
{
	t_pxobject x_obj;
	t_symbol *x_sym;
	float x_maxinc;		/* maximum inc allowed per samp */
	float x_fixdel;		/* (sr of write) / (sr of vd) */
	float x_fudge;				/* 1/(n-1) */
	long x_maxspan;				/* maximum travel allowed per call */
	float x_delonset;				/* correction to subtract from deltimes */
	float x_writevecsize;				/* number of samps in delwrite per vd call */
} t_vd;

void vd_dsp(t_vd *x, t_signal **sp, short *count);
void vd_assist(t_vd *x, void *b, long m, long a, char *s);
void *vd_new(t_symbol *s, double maxinc, long down);
t_int *vd_perform(t_int *w);
void cu_vd(float *in, float *out, t_delbuf *b, t_vd *x, long n);
short make_samp_tab(void);

void *vd_class;
t_symbol *ps_delwrite_class;
t_sampfilt *sampfilt_tab;

void main(void)
{
	setup((t_messlist **)&vd_class, (method)vd_new, (method)dsp_free, (short)sizeof(t_vd), 
		0L, A_SYM, A_DEFFLOAT, A_DEFLONG, 0);
	addmess((method)vd_dsp, "dsp", A_CANT, 0);
	addmess((method)vd_assist, "assist", A_CANT, 0);
	dsp_initclass();
	ps_delwrite_class = gensym("delwrite_class");
	make_samp_tab();
}
	
void vd_dsp(t_vd *x, t_signal **sp, short *count)
{
	t_delwrite *wt;
	
	wt = (t_delwrite *)(x->x_sym->s_thing);
	if (wt && ob_messlist(wt) == *(t_messlist **)ps_delwrite_class->s_thing) {
		x->x_delonset = wt->x_on ? 0 : sp[0]->s_n / (sp[0]->s_sr * .001);
		x->x_maxspan += ((-x->x_maxspan) & (SYS_MINBLKSIZE-1));
		x->x_fudge = (sp[0]->s_n > 1 ? 1./(sp[0]->s_n-1) : 0);
		x->x_writevecsize = sp[0]->s_n * x->x_fixdel;
		dsp_add(vd_perform, 5, x, &wt->x_buf, sp[0]->s_vec, sp[2]->s_vec, sp[0]->s_n);
	} else
		error("vd~ %s: no corresponding delwrite~",x->x_sym->s_name);
}

void vd_assist(t_vd *x, void *b, long m, long a, char *s)
{
	if (m==1) {
		switch (a) {
			case 0: sprintf(s,"(dummy) old order forcing inlet"); break;
		}
	} 
	else if (m==2) {
		switch (a) {
			case 0: sprintf(s,"(signal) delayed signal out"); break;
		}
	}
}

void *vd_new(t_symbol *s, double maxinc, long down)
{
	t_vd *x = (t_vd *)newobject(vd_class);
	dsp_setup((t_pxobject *)x,2);			// unused signal inlet
	x->x_sym = s;
	if (maxinc < 1.)
		x->x_maxinc = 4.;
	else
		x->x_maxinc = maxinc;
	x->x_maxspan = x->x_maxinc * sys_getsr();
	x->x_fudge = 1./(double)(sys_getsr()-1.);
	if (down < 0) 
		down = 0;
	else if (down > 15) 
		down = 15;
	x->x_fixdel = 1./(1 << down);
	outlet_new((t_object *)x, "signal");
	return (x);
}

t_int *vd_perform(t_int *w)
{
	float *in, *out;
	t_delbuf *b;
	t_vd *x;
	int n;
	
	x = (t_vd *)(w[1]);
	if (x->x_obj.z_disabled)
		goto out;
	b = (t_delbuf *)(w[2]);
	in = (float *)(w[3]);
	out = (float *)(w[4]);
	n = w[5];
	cu_vd(in,out,b,x,n);
out:
	return w + 6;
}

void cu_vd(float *in, float *out, t_delbuf *b, t_vd *x, long n)
{
	/* bp is incremented each sample and corresponds to zero delay;
		 at the outset it points into the buffer proper but at the end it
		 could point n-1 samps past the logical end.	Since the offset is
		 bounded below by n, bp - offset does not point past the end;
		 in fact you can read four consecutive samples starting from there
		 (but bp-offset might point before the beginning; this is guarded
		 against in the loop.) */
	float *bp = b->b_shit + b->b_phase;
	long minoffset = n * SAMP_NPOINTS;
	long maxoffset = (b->b_nsamps - 3) * SAMP_NPOINTS;
	while (n--)
	{
		float *fp;
		float foffset = *(in++) * b->b_conv - x->x_delonset;
		long bigoffset,  offset,  interpindex;
		t_sampfilt *t;
		bigoffset = foffset * SAMP_NPOINTS; 
		if (bigoffset < minoffset) bigoffset = minoffset;
		else if (bigoffset > maxoffset) bigoffset = maxoffset;
		offset = bigoffset >> SAMP_NBITS;
		interpindex = bigoffset & (SAMP_NPOINTS-1);
		t = sampfilt_tab + interpindex;
		/* printf("(in %f) %f, %d\n", in[-1], foffset, offset); */
		fp = bp - offset;
		if (fp < b->b_shit + 3) fp += b->b_nsamps;
		*out++ = fp[0] * t->f1 + fp[-1] * t->f2
		+ fp[-2] * t->f3 + fp[-3] * t->f4;
		bp++;
	}
}

short make_samp_tab(void)
{
	long i;
	
	if (SAMP_NPOINTS * sizeof(t_sampfilt) != 0x1000)
		error("make_samp_tab");
	if (!(sampfilt_tab = (t_sampfilt *)sysmem_newptr(SAMP_NPOINTS * sizeof(t_sampfilt))))
		return (0);
	for (i = 0; i < SAMP_NPOINTS; i++) {
		float f = i * (1./SAMP_NPOINTS);
		sampfilt_tab[i].f1 = -.1666667 * f * (1-f) * (2-f);
		sampfilt_tab[i].f2 = .5 * (1+f) * (1-f) * (2-f);
		sampfilt_tab[i].f3 = .5 * (1+f) * f * (2-f);
		sampfilt_tab[i].f4 = -.1666667 * (1+f) * f * (1-f);
	}
	return (1);
}
