/*
 *	Copyright (c) 1993 by IRCAM
 *	All rights reserved.
 *
 *	For any information regarding this and other IRCAM software, please
 *	send email to:
 *	manager@ircam.fr
 *
 *	$Revision: 2.6 $ IRCAM $Date: 1993/11/19 16:22:06 $
 *
 * FTS by Miller Puckette
 *
 */

/*
	jack -- the pitch~ object with partial tracking added.
*/

#include "ext.h"
#include "z_dsp.h"
#include <math.h>
#include "fft_mayer.proto.h"

//#include "m_extern.h"
//#include "d_graph.h"
//#include "d_ugen.h"
//#define __STRICT_BSD__
#include "math.h"
//#undef __STRICT_BSD__

#define UCALL_PAGE_SIZE 0
#define MAXPARTIAL 32
#define NAMPCOEF 8
#define MINPOINTS 128
#define MAXPOINTS 512
#define MAXLOG 9
#define FMAXPOINTS 512.

#define MAXVS 256
#define PEROCTAVE 24
#define FPEROCTAVE 24.
#define PERHALFTONE 2

#define NKERNEL (PEROCTAVE * MAXLOG - 2 * PEROCTAVE)
#define FREQSTEP 1.0293022
#define LOGSTEP .028881096
//#define PI 3.14159265 // defined in msp headers with more precision
//#define TWOPI (2. * PI)
#define CUTOFF .05
#define ONQUALITYTHRESH .5
#define OFFQUALITYTHRESH .25
#define ONPOWTHRESH .003
#define OFFPOWTHRESH .001
#define STICK .3
#define OVERLAP 2.

#define EXTRAKERNEL 73		/* number of extra filter kernels to keep at end */
#define ALIGN 16			/* byte alignment for microcode */
#define LOWINDEX 6
#define DEFREATTACK 1
#define DEFRATIME 100
#define DEFDEBOUNCETIME 150

#define DUMTAB1SIZE 256
#define DUMTAB2SIZE 1024
static float jacktab1[DUMTAB1SIZE], jacktab2[DUMTAB2SIZE];
#define HISTORY 16
#define POWCHUNK 64

#define S_OFF 0
#define S_ON 1
#define S_NEXTON 2

#define FTS

typedef struct kernel
{
	float *k_coef;
	int k_ncoef;
	int k_onset;
	float k_powcorrect;
} t_kernel;

typedef struct peak
{
	float p_freq;
	float p_amp;
	int p_status;
	struct peak *p_old;
} t_peak;

typedef struct sigjack
{
	//t_head x_h;
	//t_sig *x_io[IN1+OUT0];
	t_pxobject x_obj;
	void *x_clock;
	void *x_poweroutlet;
	void *x_micropitchoutlet;
	void *x_attackoutlet;
	void *x_pitchoutlet;
	void *x_partialoutlet;

	float x_power;
	float x_micropitch;
	int x_attack;

	float *x_buf;
	float *x_ebuf;
	float *x_fillbuf;
	int x_npts;
	int x_nchan;
	float x_lofreq;
	float x_reffreq;		/* frequency corresponding to MIDI 69 */
	float x_stretchcoef;		/* inverse natural log of a MIDI halftone */
	float x_onpowthresh;
	float x_offpowthresh;
	float x_onqualitythresh;
	float x_offqualitythresh;
	float x_stick;
	float x_ampcoef[NAMPCOEF];
	long x_glisstime;
	float x_halftonesperframe;
	float x_srate;
	int x_powtime;
	int x_powbuckets;
	float x_powslopethresh;
	float x_lastpow;
	int x_lopitch;
	int x_hipitch;
	int x_lopoint;
	int x_hipoint;
	int x_lastindex;
	float x_midipit;		/* pitch of ongoing note; zero if none */
	float x_pitchoutput;		/* pitch to send to outlet on clock tick */
	int x_overlap;
	float *x_freeme;
	int x_printme;
	int x_peaked;
	long x_midilastout;
	long x_miditime;
	long x_debouncetime;
	float x_maxerror;
	int x_histindex;
	float x_pitchhist[HISTORY];
	float x_amphist[HISTORY];
	int x_buckets;
	float x_vibrato;
	long x_vibtime;
	int x_countup;
	int x_loud;
	int x_maxpeaks;
	long x_downsample;
	float x_peakthresh;
	t_peak x_peaks[MAXPARTIAL];
	int x_npointsout;		/* number of points to output; 0 if lazy */
	char x_fftspace[2*sizeof(float) * MAXPOINTS + UCALL_PAGE_SIZE];	// moved here to
			// try to avoid running out of stack
} t_sigjack;

float sigjack_defcoef[] =
{ 1., .9, .8, .7, .7, .6, .6, .6};

t_kernel *pt_kernel;
float *sigjack_wind256, *sigjack_wind512;
extern int h_ishost;
void *sigjack_class;

t_symbol *ps_list;

double sigjack_mtof(double f);
double sigjack_ftom(double f);
float sigjack_getfloat(t_atom *at);
int sigjack_ilog2(int n);
double dumsqrt(double d);
void pt__init(void);
void sigjack__makespect(float *fftbuf, t_kernel *kp, float *spect, int n, int npts);
t_int *sigjack_perform(t_int *w);
void cu_vmul(float *a, float *b, float *out, long n);
void cu_vadd(float *a, float *b, float *out, long n);
void cu_sperse(float *a, float *b, long n);
void sigjack_doit(t_sigjack *x);
void sigjack_downsample(float *in, float *out, long np, long ds);
void sigjack_dsp(t_sigjack *x, t_signal **sp);
void sigjack_clear(t_sigjack *x);
void sigjack_int(t_sigjack *x, long n);
void sigjack_loudest(t_sigjack *x, long n);
void sigjack_pt_tune(t_sigjack *x, double f);
void sigjack_stretch(t_sigjack *x, double f);
void sigjack_ampl_threshold(t_sigjack *x, double f);
void sigjack_quality_threshold(t_sigjack *x, double f);
void sigjack_amp_range(t_sigjack *x, double f1, double f2);
void sigjack_quality_range(t_sigjack *x, double f1, double f2);
void sigjack_pitch_range(t_sigjack *x, long n1, long n2);
void sigjack_fixpitch(t_sigjack *x);
void sigjack_stick(t_sigjack *x, double f);
void sigjack_maxerror(t_sigjack *x, double f);
void sigjack_coefs(t_sigjack *x, t_symbol *s, int ac, t_atom *av);
void sigjack_glisstime(t_sigjack *x, long n);
void sigjack_vibrato(t_sigjack *x, long vibtime, double vibdepth);
void sigjack_debouncetime(t_sigjack *x, long n);
void sigjack_setdownsample(t_sigjack *x, long n);
void sigjack_reattack(t_sigjack *x, double f, long n);
void sigjack_peakthresh(t_sigjack *x, double f);
void sigjack_nfilters(t_sigjack *x, long npartials, long npointsout);
void sigjack_print(t_sigjack *x, long n);
void sigjack_loud(t_sigjack *x, long n);
void sigjack_tick(t_sigjack *x);
void sigjack_clean(t_sigjack *x);
void *sigjack_new(long npoints, long period, long npartials, long npointsout);
void sigjack_assist(t_sigjack *x, void *b, long m, long a, char *s);

void main(void)
{
	setup((t_messlist **)&sigjack_class, (method)sigjack_new, (method)sigjack_clean,
		(short)sizeof(t_sigjack), 0, A_DEFLONG, A_DEFLONG, A_DEFLONG, A_DEFLONG, 0);
	addmess((method)sigjack_dsp, "dsp", A_CANT, 0);
	addint((method)sigjack_int);
	addmess((method)sigjack_loudest, "loudest", A_LONG, 0);
	addmess((method)sigjack_pt_tune, "pt_tune", A_FLOAT, 0);
	addmess((method)sigjack_ampl_threshold, "ampl_threshold", A_FLOAT, 0);
	addmess((method)sigjack_quality_threshold, "quality_threshold", A_FLOAT, 0);
	addmess((method)sigjack_pt_tune, "reference-pitch", A_FLOAT, 0);
	addmess((method)sigjack_maxerror, "max-error", A_FLOAT, 0);
	addmess((method)sigjack_stretch, "stretch", A_FLOAT, 0);
	addmess((method)sigjack_amp_range, "amp-range", A_FLOAT, A_FLOAT, 0);
	addmess((method)sigjack_quality_range, "quality-range", A_FLOAT, A_FLOAT, 0);
	addmess((method)sigjack_pitch_range, "pitch-range", A_LONG, A_LONG, 0);
	addmess((method)sigjack_stick, "stick", A_FLOAT, 0);
	addmess((method)sigjack_coefs, "coefs", A_GIMME, 0);
	addmess((method)sigjack_glisstime, "gliss-time", A_LONG, 0);
	addmess((method)sigjack_vibrato, "vibrato", A_LONG, A_FLOAT, 0);
	addmess((method)sigjack_debouncetime, "debounce-time", A_LONG, 0);
	addmess((method)sigjack_reattack, "reattack", A_FLOAT, A_LONG, 0);
	addmess((method)sigjack_print, "print", A_DEFLONG, 0);
	addmess((method)sigjack_loud, "loud", A_DEFLONG, 0);
	addmess((method)sigjack_peakthresh, "peak-threshold", A_DEFFLOAT, 0);
	addmess((method)sigjack_nfilters, "nfilters", A_DEFLONG, A_DEFLONG, 0);
	addmess((method)sigjack_assist, "assist", A_CANT, 0);
	addmess((method)sigjack_setdownsample, "downsample", A_LONG, 0);
	
	ps_list = gensym("list");
	dsp_initclass();
}

void sigjack_assist(t_sigjack *x, void *b, long m, long a, char *s)
{
	if (m==1) {
		switch (a) {
			case 0: sprintf(s,"(signal) Input"); break;
		}
	} 
	else if (m==2) {
		switch (a) {
			case 0: sprintf(s,"(float) Pitch"); break;
			case 1: sprintf(s,"(bang) Attack Detected"); break;
			case 2: sprintf(s,"(float) Micropitch"); break;
			case 3: sprintf(s,"(float) Power"); break;
			case 4: sprintf(s,"(list) Partials"); break;
		}
	}
}

double sigjack_mtof(double f)
{
	return (440. * exp(.057762265 * (f - 69.)));
}

double sigjack_ftom(double f)
{
	if (f > 0) return (69. + (1./.057762265) * log(f/440.));
	else return (0);
}

float sigjack_getfloat(t_atom *at)
{
	float f;
	if (at->a_type == A_LONG) 
		f = at->a_w.w_long;
	else if (at->a_type == A_FLOAT)
		f = at->a_w.w_float;
	else { //if (sscanf(at->a_w.w_sym, "%f", &f) < 1) {
		post("warning: %s: bad number", at->a_w.w_sym->s_name);
		f = 0;
	}
	return f;
}

int sigjack_ilog2(int n)
{
	int ret = -1;
	while (n)
		{
			n >>= 1;
			ret++;
		}
	return (ret);
}

double dumsqrt(double d)
{
	float f = d;
	long l = *(long *)(&f);
	if (f < 0) return (0.);
	else return (jacktab1[(l >> 23) & 0xff] * jacktab2[(l >> 13) & 0x3ff]);
}

void pt__init(void)
{
	float *specspace, *espec;
	/* estimate total size of kernels: cover the whole spectrum 
		 4 times and overlap 4 points apiece */
	int specsize = 4 * NKERNEL + OVERLAP * MAXPOINTS, filtno;
	int i;
	float *fftbuf, filtfreq, filtomega;
	t_kernel *kp;
	char fftspace[2*sizeof(float) * MAXPOINTS + UCALL_PAGE_SIZE], *foo;
	if (pt_kernel) return;

	if (!(specspace = (float *)t_getbytes(specsize * sizeof(*specspace))))
	{
		error("jack~: can't allocate space");
		return;
	}
	espec = specspace + specsize;
	if (!(pt_kernel = (t_kernel *)t_getbytes(NKERNEL * sizeof(*pt_kernel))))
	{
		t_freebytes(specspace, (specsize * sizeof(*specspace)));
		error("jack~: can't allocate space");
		return;
	}
	if (!(foo = (char *)t_getbytes((512+256)*sizeof(float))))
	{
		t_freebytes(specspace, (specsize * sizeof(*specspace)));
		t_freebytes(pt_kernel, NKERNEL * sizeof(*pt_kernel));
		error("jack~: can't allocate space");
		return;
	}
	sigjack_wind256 = (float *)foo;
	sigjack_wind512 = sigjack_wind256 + 256;
		/* make 256-point and 512-point hanning windows for power calc */
	for (i = 0; i < 256; i++) sigjack_wind256[i] =
		(1./256.) * (1. - cos((2 * 3.14159/256.) * i));
	for (i = 0; i < 512; i++) sigjack_wind512[i] =
		(1./512.) * (1. - cos((2 * 3.14159/512.) * i));
	/* get a page-aligned buffer for doing FFT */
	//fftbuf = (float *)(((long)fftspace) +
	//			 ((-(long)fftspace) & (UCALL_PAGE_SIZE - 1)));
	fftbuf = (float *)fftspace;

	post("fftbuf %lx fftspace %lx",fftbuf,fftspace);
	
	for (i = 0, filtfreq = 2. / FMAXPOINTS, kp = pt_kernel; i < NKERNEL; 
			 i++, filtfreq *= FREQSTEP, kp++)
	{
		float protowind = 1. / (OVERLAP * (FREQSTEP - 1.) * filtfreq),
		window, phase, wphase,
		filtomega = filtfreq * TWOPI, windomega, cutval, norm,
		midipit;
		float *fp1, *fp2, *fp3, *fp4;
		int nbin;
		int j;
		window = (protowind > .5 * FMAXPOINTS ? .5 * FMAXPOINTS : protowind);
		windomega = PI / window;
		set_zero(fftbuf, 2 * MAXPOINTS);
		norm = 1./window;
		for (fp1 = fp2 = fftbuf + (MAXPOINTS/2), 
			fp3 = fp4 = fftbuf + MAXPOINTS + (MAXPOINTS/2), j = window,
			phase = 0., wphase = 0.; j--;
		 	fp1--, fp3--, fp2++, fp4++, phase += filtomega, wphase += windomega)
		{
			double wind = norm * (.5 + .5 * cos(wphase));
			fp1[0] = fp2[0] =	cos(phase) * wind;
			fp3[0] = -(fp4[0] =	sin(phase) * wind);
		}
		fft(MAXPOINTS,fftbuf, fftbuf + MAXPOINTS);
		
		//// YOU"RE HERE
		cutval = CUTOFF;
		for (j = 0, fp2 = fftbuf + (MAXPOINTS/2); j < (MAXPOINTS/2); j++, fp2--)
			if (*fp2 >= cutval || *fp2 < (-cutval))
				break;
		for (j = 0, fp1 = fftbuf; j < (MAXPOINTS/2); j++, fp1++)
			if (*fp1 >= cutval || *fp1 < (-cutval))
				break;
		if (fp1 >= fp2) {
			static long bugged = 0;
			if (!bugged) {
				error("jack~ %d: zero-length kernel", i);
				bugged = 1;
			}
			fp1 = fp2;
		}
		nbin = 1 + ((fp2 - fp1));
			/* if (i == 100 || i == 101)
	 {
	 float *fp3;
	 int z = 10;
	 post("n = %d; start at %d", nbin, (fp1 - fftbuf) >> 1);
	 for (z = 10, fp3 = fp1; fp3 <= fp2 && (z -- > 0); fp3 += 2)
	 post("%f %f", fp3[0], fp3[1]);
	 post("(from zero:");
	 for (z = 10, fp3 = fftbuf; z -- > 0; fp3 += 2)
	 post("%f %f", fp3[0], fp3[1]);
	 } */
		if (specspace + nbin >= espec)
		{
			post("bug: jack~: misestimated the number of bins");
			return;
		}
			
		kp->k_onset = fp1 - fftbuf;
		kp->k_coef = specspace;
		kp->k_ncoef = nbin;
		kp->k_powcorrect = window / protowind;
			/* post("%d\n", nbin); */
		for (; nbin--; fp1 ++)
			*specspace++ = *fp1;
	}
	
	for (i = 0; i < DUMTAB1SIZE; i++)
	{
		float f;
		long l = (i ? (i == DUMTAB1SIZE-1 ? DUMTAB1SIZE-2 : i) : 1)<< 23;
		*(long *)(&f) = l;
		jacktab1[i] = sqrt(f);	
	}
	for (i = 0; i < DUMTAB2SIZE; i++)
	{
		float f = 1 + (1./DUMTAB2SIZE) * i;
		jacktab2[i] = sqrt(f);	
	}
}

void sigjack__makespect(float *fftbuf, t_kernel *kp, float *spect, int n, int npts)
{
	while (n--)
	{
		float rsum = 0, isum = 0;
		float *rfp = fftbuf + kp->k_onset;
		float *ifp = fftbuf + npts + kp->k_onset;
		float *cp = kp->k_coef;
		int k = kp->k_ncoef;
		while (k--)
		{
			float coef = *cp++;
			rsum += *rfp++ * coef;
			isum += *ifp++ * coef;
		}
		*spect++ = rsum * rsum + isum * isum;
		kp++;
	}
}

t_int *sigjack_perform(t_int *w)
{
	t_sigjack *x;
	float *sig;
	long n;
	float *fp;
	
	x = (t_sigjack *)(w[1]);
	sig = (float *)(w[2]);
	n = w[3];
	
	if (x->x_obj.z_disabled)
		goto out;

	fp = x->x_fillbuf;

	sigjack_downsample(sig, fp, n, x->x_downsample);
	fp += (n / x->x_downsample);
	if (fp != x->x_ebuf)
		x->x_fillbuf = fp;
	else
		sigjack_doit(x);

out:
	return w + 4;
}

void cu_vmul(float *a, float *b, float *out, long n)
{
	while (n--)
		*out++ = *a++ * *b++;
}

void cu_vadd(float *a, float *b, float *out, long n)
{
	while (n--)
		*out++ = *a++ + *b++;
}

void cu_sperse(float *a, float *b, long n)
{
	float temp[MAXPOINTS];
	long i,j;
	
	for (i = 0; i < n; i++)
		temp[i] = a[i];
	
	for (i = 0, j = 0; i < n; i++, j+= 2) {
		a[j] = temp[i];
		a[j+1] = b[i];
	}
}

void sigjack_doit(t_sigjack *x)
{
	int npts = x->x_npts;
	int nchan = x->x_nchan, index = -1, i, j, stable,
		lo = x->x_lopoint-2, hi = x->x_hipoint;
	float bestamp, pitchpow, totpow, mspow, newpitch,
		*fp, *fp2, *fp3, *fftbuf;
	t_kernel *kp = pt_kernel;
	
	//char fftspace[sizeof(float) * (2*MAXPOINTS) + UCALL_PAGE_SIZE];
	float para[3], correct, spectrum[NKERNEL + EXTRAKERNEL],
		pitches[NKERNEL], ampspect[NKERNEL + EXTRAKERNEL], quality;
	t_peak newpeaks[MAXPARTIAL];
	int nnewpeaks = 0;

		/* get a page-aligned buffer for doing FFT */
	//fftbuf = (float *)(((long)fftspace) +
	//			 ((-(long)fftspace) & (UCALL_PAGE_SIZE - 1)));
	fftbuf = (float *)x->x_fftspace;
	fp3 = fftbuf + POWCHUNK;
	set_zero(fftbuf, 2 * POWCHUNK);	/* clear and warm up cache */
	
		/* get mean-square power */
	for (fp = (x->x_npts == 512 ? sigjack_wind512 : sigjack_wind256),
		fp2 = x->x_buf, i = npts; i -= POWCHUNK;
			fp += POWCHUNK, fp2 += POWCHUNK)
	{
		cu_vmul(fp2, fp2, fp3, POWCHUNK);
		cu_vmul(fp, fp3, fp3, POWCHUNK);
		cu_vadd(fp3, fftbuf, fftbuf, POWCHUNK);
	}
	for(i = POWCHUNK, fp = fftbuf, mspow=0; i--;)
		mspow += *fp++;
	x->x_power = 100 + 8.6858 *log(sqrt(mspow) + 1e-20);

	set_zero(fftbuf, 2 * npts);		/* zero it out */
	//sigjack_mux(x->x_buf, fftbuf, 2, npts);	/* copy input into real part */
	sysmem_copyptr(x->x_buf, fftbuf, npts * sizeof(float));
	x->x_fillbuf = x->x_buf + x->x_overlap;
	if (x->x_overlap)
		sysmem_copyptr(x->x_buf + (npts - x->x_overlap), x->x_buf,
					x->x_overlap * sizeof(float));
	
	//fft(fftbuf, npts, 0);
	fft(npts,fftbuf,fftbuf + npts);
	/* get a bounded-Q power spectrum, quarter-tone spaced */
	sigjack__makespect(fftbuf, pt_kernel, spectrum, x->x_nchan,npts);
	for (i = EXTRAKERNEL, fp = spectrum + NKERNEL; i--;)
		*fp++ = 0;
	
	if (x->x_printme >= 2)
	{
		int z;
		float *fp3 = spectrum;
		post("raw filter output:");
		for (z = 0; z < nchan; z += 4, fp3 += 4)
			post("%3.3d %12.1f %12.1f %12.1f %12.1f",
		 		z, fp3[0], fp3[1], fp3[2], fp3[3]);
	}
	
	/* make an amplitude spectrum */
	for (i = NKERNEL, fp = spectrum, fp2 = ampspect; i--;)
//#ifdef I860
		*fp2++ = dumsqrt(*fp++);
//#else I860
//	*fp2++ = sqrt(*fp++);
//#endif I860
	for (i = EXTRAKERNEL; i--;) *fp2++ = 0;
	
	
	/* look for bin with max. power in (itself and first few
		 overtones.) */
	for (i = 0, fp = ampspect, fp2 = x->x_ampcoef, bestamp=0;
			 i < nchan; fp++, i++)
	{
			float mofo = fp[0] * fp2[0] + fp[24] * fp2[1] + fp[38] * fp2[2] +
				fp[48] * fp2[3] + fp[56] * fp2[4] + fp[62] * fp2[5] + fp[67] * fp2[6] +
				fp[72] * fp2[7];

			pitches[i] = mofo;
			if (i >= lo && i < hi && mofo > bestamp)
				bestamp = mofo, index = i;
	}
	if (index <= 0)			/* didn't find anything at all? */
	{
		newpitch = 0;
		x->x_lastindex = 0;
		goto done;
	}
	
	if (x->x_lastindex)		/* possibly stick to previous pitch */
	{
		int lastindex = x->x_lastindex;
		float x2;
		fp = pitches + lastindex;
		x2 = *fp;
		if (fp[-1] > x2 && lastindex > lo)
			x2 = *--fp, lastindex--;
		else if (fp[1] > x2 && lastindex < hi - 1)
			x2 = *++fp, lastindex++;
		if (x2 > (1. - x->x_stick) * bestamp)
			index = lastindex;
	}
	
	/* get total power for the good overtones */
	fp = spectrum + index;
	pitchpow = fp[0] + fp[24] + fp[38] + fp[48] + fp[56] +
		fp[62] + fp[67] + fp[72];
	
	/* get the total power from DC up to the 8th harmonic.
		 The k_powcorrect factor is to correct for the fact that the
		 filterbank overlap is greater in the low bands and so
		 merely summing them would weight them too highly. */
	fp += 72;
	if (fp >= spectrum + NKERNEL)
		fp = spectrum + (NKERNEL - 1);
	for (fp2 = spectrum, totpow = 0., kp = pt_kernel; fp2 < fp; fp2++, kp++)
		totpow += *fp2 * kp->k_powcorrect;
	
	if (x->x_printme)
		post("index %d, value %f; overall %f", index,
			(float)sqrt(pitchpow)/npts, (float)sqrt(totpow)/npts);
	
	quality = pitchpow / totpow;

	/* to report a pitch there must be at least "offqualitythresh"
		 times the total power in the good bins. */
	if (pitchpow < x->x_offpowthresh || quality < x->x_offqualitythresh)
	{
		newpitch = 0;
		x->x_lastindex = 0;
		if (x->x_printme)
			post("no pitch (q or power too low)");
		goto done;
	}

	fp = spectrum + index;

	/* remember the index for pitch-stick test next time */
	if (pitchpow > x->x_onpowthresh)
		x->x_lastindex = index;
	else
		x->x_lastindex = 0;
	
	/* do a parabolic correction to increase precision */
	for (i = 0, fp = spectrum + index - 1; i < 3; i++, fp++)
		para[i] = fp[0] + 2.*fp[24] + 3.*fp[38] + 4.*fp[48] +
			5.*fp[56] + 6.*fp[62] + 7.*fp[67] + 8.*fp[72];
	if (para[0] >= para[1])
		correct = -.5;
	else if (para[2] >= para[1])
		correct = .5;
	else
	{
		double f1 = 1. - para[0]/para[1], f2 = 1. - para[2]/para[1];
		correct = (f1-f2)/(2.*(f1+f2) - 3.*f1*f2);
	}
	/* correct = .5 * (para[2] - para[0])
		 / (2. * para[1] - para[0] - para[2]); */
	
	newpitch = 69. + (12./log(2.)) * (
		log(x->x_lofreq/x->x_reffreq) +
			x->x_stretchcoef * LOGSTEP * (index + correct));
			
	// correct for downsampling
	//if (x->x_downsample == 2)
	//	newpitch -= 12.;
	//else if (x->x_downsample == 4)
	//	newpitch -= 24.;
		
	stable = (x->x_glisstime == 0 ||
			 (newpitch >= x->x_micropitch - x->x_halftonesperframe &&
				newpitch <= x->x_micropitch + x->x_halftonesperframe));
	if (x->x_printme)
	{
		post("pitch %f, %sstable", newpitch, (stable ? "" : "un"));
	}
done:
	x->x_histindex = (x->x_histindex + 1)&(HISTORY-1);
	x->x_pitchhist[x->x_histindex] = newpitch;
	x->x_amphist[x->x_histindex] = mspow;
	if (x->x_midipit > 0)
	{
		float error;
		/* test for turnoff -- if "newpitch" is more than x->x_vibrato
	of a halftone away from the current note, (or from an
	octave up or down from it) the current note has stopped
	playing.	This also takes care of the case where
	newpitch == 0. */
		error = newpitch - x->x_midipit;
		if (error > 6)
			error -= 12;
		else if (error < -6)
			error += 12;
		if (error > x->x_vibrato || error < -x->x_vibrato)
		{
			x->x_midipit = 0;
				/* bash previous pitch history.	This prevents the stability criterion
	from being met until x->x_buckets new pitches have been measured. */
			x->x_pitchhist[(x->x_histindex - 1)&(HISTORY-1)] = 0;
		}
	 }
		/* now test for reattack, which is a condition on the
			amplitude envelope.	Reattack is also marked by setting current
	pitch to zero; the note-on conditions must apply before a new
	note is output. */
	if (!x->x_peaked)
	{
		if (x->x_powbuckets &&
			(mspow * (x->x_powslopethresh + 1) < 
				x->x_amphist[(x->x_histindex - x->x_powbuckets) & (HISTORY-1)]))
			x->x_peaked = 1;
	}
	else
	{
		if (x->x_powbuckets && pitchpow > x->x_onpowthresh
			&& (mspow > (x->x_powslopethresh + 1) *
				x->x_amphist[(x->x_histindex - x->x_powbuckets) & (HISTORY-1)]))
		{
			x->x_attack = 1;
			x->x_midipit = 0;
			if (x->x_loud)
				post("reattack: power %f", pitchpow);
			x->x_peaked = 0;
	/* here's the good part -- if there's a reattack and we haven't
	reported a note yet since the last one, look again RETROACTIVELY,
	using laxer criteria */
			if (x->x_countup < 100)
			{
				float bestpow = 0, bestpitch, recentpow = mspow;
				int histdepth = (x->x_countup > HISTORY ? HISTORY : x->x_countup),
				gobackup = 0;
				for (i = 1; i < histdepth; i++)
				{
					float npow = x->x_amphist[(x->x_histindex - i) & (HISTORY-1)];
					if (!gobackup)
					{
						if (npow > recentpow)
							gobackup = 1;
						recentpow = npow;
					}
					else if (npow > bestpow)
					{
						bestpow = npow;
						bestpitch = x->x_pitchhist[(x->x_histindex - i) & (HISTORY-1)];
					}
				}
				if (x->x_loud)
					post("a posteriori: countup=%d, bestpow = %f; pitch = %f",
						x->x_countup, bestpow, bestpitch);
				if (bestpow > 0)
					x->x_midipit = x->x_pitchoutput = bestpitch;
			}
			x->x_countup = 0;
		}
	}
	x->x_countup++;
	x->x_lastpow = mspow;
	
	/* test for turnon.	We must: not have a current pitch;
		 have a candidate pitch; have at least a certain
		 amount of power and "quality". */
	if ((x->x_midipit == 0) && newpitch != 0 &&
			pitchpow > x->x_onpowthresh &&
			quality > x->x_onqualitythresh)
	{
		int ipit, deviated = 0;
		float error, pitsum;
			/* in addition we must have a stable pitch in the sense that it
			does not deviate more than x->x_vibrato from the mean over the
			last x->x_buckets values. */
	
		ipit = newpitch + .5;
		error = newpitch - ipit;
		if (x->x_buckets)
		{
			for (pitsum = 0, i = 0; i < x->x_buckets; i++)
				pitsum += x->x_pitchhist[(x->x_histindex-i)&(HISTORY-1)];
			pitsum /= x->x_buckets;
			for (i = 0; i < x->x_buckets; i++)
			{
				float deviation =
					x->x_pitchhist[(x->x_histindex-i)&(HISTORY-1)] - pitsum;
				if (x->x_printme)
					post("deviation %f", deviation);

				if (deviation < -x->x_vibrato || deviation > x->x_vibrato)
				{
					if (x->x_printme)
						post("called off");
					deviated = 1;
				}
			}
		}
		else
			pitsum = newpitch;
		
		if ((error < x->x_maxerror) && (error > -x->x_maxerror) && !deviated)
		{
			int rats = x->x_midilastout;
			if (gettime() > x->x_miditime ||
				(ipit != rats && ipit != rats + 12 && ipit != rats - 12))
			{
				if (x->x_loud)
				{
					post("ongoing note: pitch %f; power %f; quality %f",
						pitsum, x->x_power, quality);
					if (x->x_buckets)
						post("stable over %d points", x->x_buckets);
					else
						post("no stability criterion");
				}
				x->x_miditime = gettime() + x->x_debouncetime;
				x->x_midilastout = ipit;
				x->x_peaked = 0;
				x->x_midipit = x->x_pitchoutput = pitsum;
				x->x_countup = 1000;
			}
		}
		else if (x->x_printme)
			post("out of tune");
	}
	x->x_micropitch = newpitch;
	
/* -------------- track individual harmonics ---------------/

	
	/* square everything in the (first half of the) FFT buffer;
		put result in the second half */
	//set_zero(fftbuf+npts, npts);	/* warm up cache */
	// ddz: in order not to rewrite this algorithm, I intersperse
	// real and imaginary in the first half of the buffer
	cu_sperse(fftbuf,fftbuf + npts,npts/2);
	cu_vmul(fftbuf, fftbuf, fftbuf+npts, npts);	/* square the points */
	nnewpeaks = 0;
	if (totpow > x->x_offpowthresh)
	{
		float last = 0, prev, new = 0, thresh = (2./MAXPARTIAL);
		float peakamp, peakfreq;
		float tpiovern = 2*3.14159/(float)npts;
			/* Normalizing factor for amplitude.	The 0.088... is a fudge. */
		float normalizer = 0.088269698 * sqrt(1./(float)npts);
		int nprint = 3, k;
		fp = fftbuf+6;
		fp2 = fp+npts;
		new = fp2[-2] + fp2[-1];
		last = fp2[-4] + fp2[-3];
		for (i = (npts>>1)-5, j = 2; i--; fp += 2, fp2 += 2, j++)
		{
			float det, threshtimes, hanningbuf[6];
			prev = last;
			last = new;
			new = fp2[0] + fp2[1];
			if (last > new && last > prev) /* candidate: local maximum in power.	*/
			{
				float det, threshtimes, hanningbuf[6], hanningpow[3], *fp4;

 				nprint--;
 				if (x->x_printme && (nprint > 0))
				{
					post("index %d prev %f last %f new %f",
						j, prev, last, new);
				}

				/* calculate Hanning windowed FFT for 3 pts surrounding the peak */

				for (k = 6, fp3 = fp-6, fp4 = hanningbuf; k--; fp3++, fp4++)
					*fp4 = fp3[2] - .5 * (fp3[0] + fp3[4]);

				hanningpow[0] = hanningbuf[0] * hanningbuf[0] +
					hanningbuf[1] * hanningbuf[1];
				hanningpow[1] = hanningbuf[2] * hanningbuf[2] +
					hanningbuf[3] * hanningbuf[3];
				hanningpow[2] = hanningbuf[4] * hanningbuf[4] +
					hanningbuf[5] * hanningbuf[5];


				/* power test -- compare hanningpow[1] to mspow */
				
				if (hanningpow[1] * x->x_maxpeaks * x->x_peakthresh < npts * mspow)
					continue;
 				if (x->x_printme && (nprint > 0))
				{
					post("power %f %d %f %f", hanningpow[1], x->x_maxpeaks,
						npts * mspow, x->x_peakthresh);
				}
		

				/* dot product test */
		
				if (hanningbuf[0] * hanningbuf[2] + hanningbuf[1] * hanningbuf[3] > 0)
					continue;
				if (hanningbuf[4] * hanningbuf[2] + hanningbuf[5] * hanningbuf[3] > 0)
					continue;

				if (x->x_printme && (nprint > 0))
					post("--------- good -----------");

		/* Measure the "true" amplitude and frequency.
		First we get power which is approximated as a linear combination
		of the powers of the three closest bins, Hanning windowed; the
		linear combination was chosen by considering the theoretical
		cases of a partial in the center of a bin and the case of one
		halfway between two bins.	This appears to be good to 0.2% .	*/

				peakamp = normalizer * sqrt(.640823357 * hanningpow[1] +
					.71835328 * (hanningpow[0] + hanningpow[2]));

				/* frequency measurement is out of Puckette and Brown
					(unpublished) */
				{
					static float realtwiddle = 0., imtwiddle = 0.;
					float realhanning1, imhanning1, realhanning2, imhanning2,
						realprod, improd;
					if (realtwiddle == 0.)
					{
						realtwiddle = cos(tpiovern);
						imtwiddle = sin(tpiovern);
					}
					realhanning1 = hanningbuf[2];
					imhanning1 = hanningbuf[3];
					realhanning2 = fp[-2] - 0.5 *
						(realtwiddle * (fp[-4] + fp[0]) +
						imtwiddle * (fp[-3] - fp[1]));
					imhanning2 = fp[-1] - 0.5 *
						(realtwiddle * (fp[-3] + fp[1]) +
						imtwiddle * (fp[0] - fp[-4]));
					if (x->x_printme && (nprint > 0))
						post("real1=%f, im1=%f", realhanning1, imhanning1),
						post("real2=%f, im2=%f", realhanning2, imhanning2);
					realprod = realhanning1*realhanning2 + imhanning1 * imhanning2;
					improd = realhanning1*imhanning2 - realhanning2*imhanning1;
					if (realprod < 0 || improd > .1*realprod || -improd > .1*realprod)
					{
						if (x->x_printme)
							post("throwing out bad peak!");
						continue;
					}
					peakfreq = x->x_srate * (1/(2*3.14159)) * 
						(tpiovern * ((float)j) + improd/realprod);
				}
				newpeaks[nnewpeaks].p_freq = peakfreq;
				newpeaks[nnewpeaks].p_amp = peakamp;
				newpeaks[nnewpeaks].p_status = 1;
 				if (x->x_printme && (nprint > 0))
					post("amp %f; freq %f", peakamp, peakfreq);
				if (++nnewpeaks >= x->x_maxpeaks) break;
			}
		}
	}
	
		/* if x_npointsout is positive, this means we are to report that
		number of continuous harmonics, out of the (possibly greater number
		of) peaks we have found. */
		
	if (x->x_npointsout)
	{
		// I moved this stuff that was here to ifdef NPOINTSOUT0 below
		for (i = 0; i < nnewpeaks; i++)
		{
			float bin = x->x_srate / x->x_npts;
			t_peak *p1 = &newpeaks[i];
			t_peak *p2 = x->x_peaks;
			p1->p_status = -1;
			for (j = 0; j < x->x_maxpeaks; j++, p2++)
			{
				if (p2->p_status && p2->p_freq < p1->p_freq + bin &&
					p2->p_freq > p1->p_freq - bin)
				{
					p1->p_status = j;
					break;
				}
			}
		}
	
			/* initialize current states to zero */
		for (i = 0; i < x->x_maxpeaks; i++)
		{
			x->x_peaks[i].p_status = 0;
		}
	
			/* mark states of current formants that match a new one as ON */
		for (i = 0; i < nnewpeaks; i++) if (newpeaks[i].p_status >= 0)
			x->x_peaks[newpeaks[i].p_status].p_status = S_ON;
	
		/* for new formants that didn't get a match, find a free
		one and mark it "NEXTON" */
		
		for (i = 0, j = 0; i < nnewpeaks; i++)
		{
			if (newpeaks[i].p_status < 0)
			{
				while (x->x_peaks[j].p_status)
				{
					j++;
					if (j >= x->x_maxpeaks)
					{
						j = 0;
						break;
					}
				}
				newpeaks[i].p_status = j;
				x->x_peaks[j].p_status = S_NEXTON;
			}
		}
		
		/* copy new ones in.	Old ones that aren't copied onto are already
				marked "off".	*/
				
		for (i = 0; i < nnewpeaks; i++) 
		{
			t_peak *p1 = &newpeaks[i];
			t_peak *p2 = x->x_peaks + p1->p_status;
			if (p1->p_status >= 0)
			{
				p2->p_freq = p1->p_freq;
				p2->p_amp = p1->p_amp;
			}
			else
				post("jack~: bug 2");
		}
	}
	else
	{
		t_peak *p1, *p2;

		/* if x_npointsout is zero, we only produce output on demand.	Here we
		just copy the vector of peaks into the instance structure; the rest happens
		in the "int" method. */

		for (p1 = newpeaks, p2 = x->x_peaks, i = 0; i < nnewpeaks; p1++, p2++, i++)
		{
			p2->p_status = S_ON;
			p2->p_amp = p1->p_amp;
			p2->p_freq = p1->p_freq;
		}
		while (i++ < x->x_maxpeaks)
			(p2++)->p_status = 0;
	}
	x->x_printme = 0;
	clock_delay(x->x_clock, 0L);
}

void sigjack_downsample(float *in, float *out, long np, long ds)
{
	long i;
	
	for (i = 0; i < np; i+=ds, in += ds) {
		*out++ = *in;
	}
}

void sigjack_dsp(t_sigjack *x, t_signal **sp)
{
	x->x_srate = sp[0]->s_sr;
	sigjack_fixpitch(x);
	sigjack_glisstime(x, x->x_glisstime);
	sigjack_vibrato(x, x->x_vibtime, x->x_vibrato);
	sigjack_reattack(x, x->x_powslopethresh, x->x_powtime);
	dsp_add(sigjack_perform, 3, x, sp[0]->s_vec, sp[0]->s_n);
}

void sigjack_clear(t_sigjack *x)
{
	int i;
	
	for (i = 0; i < x->x_maxpeaks; i++)
		x->x_peaks[i].p_status = 0;
}

void sigjack_int(t_sigjack *x, long n)
{
	t_peak *p1, *p2;
	int nleft;
	int nsent = 0;
	t_atom form[3];
	form[0].a_type = A_LONG;
	form[1].a_type = form[2].a_type = A_FLOAT;
	
	if (x->x_npointsout)
	{
		post("sigjack_int: only works if npointsout is zero");
		return;
	}
	if (n > x->x_maxpeaks)
		n = x->x_maxpeaks;
	else if (n < 0)
		n = 0;
	for (p1 = x->x_peaks; n--; p1++)
	{
		form[0].a_w.w_long = nsent++;
		if (p1->p_status)
		{
			form[1].a_w.w_float = p1->p_amp;
			form[2].a_w.w_float = p1->p_freq;
		}
		else form[1].a_w.w_float = form[2].a_w.w_float = 0;
		outlet_list(x->x_partialoutlet, ps_list, 3, form);
	}
}

void sigjack_loudest(t_sigjack *x, long n)
{
	t_peak *p1, *p2;
	int nleft;
	int nsent = 0;
	t_atom form[3];
	form[0].a_type = A_LONG;
	form[1].a_type = form[2].a_type = A_FLOAT;
	
	if (x->x_npointsout)
	{
		post("sigjack_loudest: only works if npointsout is zero");
		return;
	}
	if (n > x->x_maxpeaks)
		n = x->x_maxpeaks;
	while (n > 0) 
	{
		float best = 0;
		p2 = 0;
		for (p1 = x->x_peaks, nleft = x->x_maxpeaks; nleft--; p1++)
		{
			if (p1->p_status == S_ON && p1->p_amp > best)
			{
				best = p1->p_amp;
				p2 = p1;
			}
		}
		if (!p2)
			return;
		form[0].a_w.w_long = nsent++;
		form[1].a_w.w_float = p2->p_amp;
		form[2].a_w.w_float = p2->p_freq;
		outlet_list(x->x_partialoutlet, ps_list, 3, form);
		p2->p_status = S_NEXTON;
		n--;
	}
	for (p1 = x->x_peaks, nleft = x->x_maxpeaks; nleft--; p1++)
		if (p1->p_status)
			p1->p_status = S_ON;
}

void sigjack_pt_tune(t_sigjack *x, double f)
{
	if (f >= 1.)
		x->x_reffreq = sigjack_mtof(f);
	else
		x->x_reffreq = 440.;
}

void sigjack_stretch(t_sigjack *x, double f)
{
	if (f < .01)
		f = 1;
	x->x_stretchcoef = f;
}

void sigjack_ampl_threshold(t_sigjack *x, double f)
{
	post("warning: ampl_threshold: does nothing; use amp-range");
}

void sigjack_quality_threshold(t_sigjack *x, double f)
{
	post("warning: quality_threshold is obsolete; use quality-range");
	x->x_onqualitythresh = f;
	x->x_offqualitythresh = .5*f;
}

void sigjack_amp_range(t_sigjack *x, double f1, double f2)
{
	double scale = x->x_npts * x->x_npts;
	x->x_offpowthresh = f1 * f1 * scale;
	x->x_onpowthresh = f2 * f2 * scale;
}

void sigjack_quality_range(t_sigjack *x, double f1, double f2)
{
	x->x_offqualitythresh = f1 * f1;
	x->x_onqualitythresh = f2 * f2;
}

void sigjack_pitch_range(t_sigjack *x, long n1, long n2)
{
	x->x_lopitch = n1;
	x->x_hipitch = n2;
	sigjack_fixpitch(x);
}

void sigjack_fixpitch(t_sigjack *x)
{
	double pitchoffirst =
		sigjack_ftom(x->x_lofreq = 2. * x->x_srate / x->x_npts);
	int offset1 = .5 + 2 * (x->x_lopitch - pitchoffirst);
	int offset2 = .5 + 2 * (x->x_hipitch - pitchoffirst);
	if (offset1 < LOWINDEX)
		offset1 = LOWINDEX;
	if (offset1 >= x->x_nchan)
		offset1 = x->x_nchan-1;
	if (offset2 < offset1)
		offset2 = offset1;
	if (offset2 >= x->x_nchan)
		offset2 = x->x_nchan-1;
	x->x_lopoint = offset1;
	x->x_hipoint = offset2;
}

void sigjack_stick(t_sigjack *x, double f)
{
	if (f < 0.) 
		f = 0.;
	else if (f > 1.)
		f = 1.;
	x->x_stick = f;
}

void sigjack_maxerror(t_sigjack *x, double f)
{
	if (f < 0.)
		f = 0.;
	else if (f > .5)
		f = .5;
	x->x_maxerror = f;
}

void sigjack_coefs(t_sigjack *x, t_symbol *s, int ac, t_atom *av)
{
	if (ac < NAMPCOEF)
		error("sigjack_coefs; need %d values", NAMPCOEF);
	else {
		int i;
		for (i = 0; i < NAMPCOEF; i++)
		{
			float f = sigjack_getfloat(av+i);
			if (f < 0)
				f = 0;
			else if (f > 1)
				f = 1;
			x->x_ampcoef[i] = f;
		}
	}
}

void sigjack_glisstime(t_sigjack *x, long n)
{
	if (n > 0)
	{
		x->x_glisstime = n;
		x->x_halftonesperframe =
			(x->x_npts - x->x_overlap)/(.001 * n * x->x_srate);
	}
	else
		x->x_glisstime = 0;
}

void sigjack_vibrato(t_sigjack *x, long vibtime, double vibdepth)
{
	if (vibdepth < 0)
		vibdepth = 0;
	if (vibtime < 0)
		vibtime = 0;
	x->x_vibtime = vibtime;
	x->x_buckets = (vibtime * .001 * x->x_srate)/(x->x_npts - x->x_overlap);
	if (x->x_buckets >= HISTORY)
		x->x_buckets = HISTORY-1;
	else if (x->x_buckets < 2)
		x->x_buckets = 0;
	x->x_vibrato = vibdepth;
}

void sigjack_debouncetime(t_sigjack *x, long n)
{
	x->x_debouncetime = (n > 0 ? n : 0);
}

void sigjack_setdownsample(t_sigjack *x, long n)
{
	if (n == 1 || n == 2 || n == 4)
		x->x_downsample = n;
	//post("downsample is now %ld",x->x_downsample);
}

void sigjack_reattack(t_sigjack *x, double f, long n)
{
	if (f > 0 && n > 0)
	{
		x->x_powslopethresh = f;
		x->x_powtime = (n < 1 ? 1 : n);
		x->x_powbuckets =
			(n * .001 * x->x_srate)/(x->x_npts - x->x_overlap);
		if (x->x_powbuckets >= HISTORY)
			x->x_powbuckets = HISTORY-1;
	}
	else
	{
		x->x_powslopethresh = 0;
		x->x_powtime = 0;
		x->x_powbuckets = 0;
	}
}

void sigjack_peakthresh(t_sigjack *x, double f)
{
	if (f < 0.)
		f = 0.;
	else if (f > 1.)
		f = 1.;
	x->x_peakthresh = f;
}

void sigjack_nfilters(t_sigjack *x, long npartials, long npointsout)
{
	if (npartials < 0)
		npartials = 0;
	else if (npartials > MAXPARTIAL)
		npartials = MAXPARTIAL;
	x->x_maxpeaks = npartials;
	if (npointsout < 0)
		npointsout = 0;
	else if (npointsout > npartials)
		npointsout = npartials;
	x->x_npointsout = npointsout;
}

void sigjack_print(t_sigjack *x, long n)
{
	if (n < 1)
		n = 1;
	x->x_printme = n;
	post("jack~:");
	post("loud %d", x->x_loud);
	post("reference-pitch %f", sigjack_ftom(x->x_reffreq));
	post("stretch %f", x->x_stretchcoef);
	post("stick %f", x->x_stick);
	post("gliss-time %d", x->x_glisstime);
	post("max-error %f", x->x_maxerror);
	post("debounce-time %d", x->x_debouncetime);
	post("reattack %f %d", x->x_powslopethresh, x->x_powtime);
	post("vibrato %ld %f", x->x_vibtime, x->x_vibrato);
	post("coefs %f %f %f %f",
			 x->x_ampcoef[0],x->x_ampcoef[1],x->x_ampcoef[2],x->x_ampcoef[3]); 
	post("...	 %f %f %f %f",
			 x->x_ampcoef[4],x->x_ampcoef[5],x->x_ampcoef[6],x->x_ampcoef[7]); 
	post("amp-range %f %f",
			 sqrt(x->x_offpowthresh)/x->x_npts,
			 sqrt(x->x_onpowthresh)/x->x_npts);
	post("quality-range %f %f", sqrt(x->x_offqualitythresh),
			 sqrt(x->x_onqualitythresh));
	post("pitch-range %d %d", x->x_lopitch, x->x_hipitch);
	if (x->x_lofreq > 1.)
	{
		post("(possible pitch range: %d - %d)",
			(int)(sigjack_ftom(x->x_lofreq) + (.5 + LOWINDEX/2)),
			(int)(sigjack_ftom(x->x_lofreq) + (.5 + x->x_nchan/2)));
	}
	post("peak-threshold %f", x->x_peakthresh);
	post("nfilters %ld %ld", x->x_maxpeaks, x->x_npointsout);
}

void sigjack_loud(t_sigjack *x, long n)
{
	x->x_loud = n;
}

void sigjack_tick(t_sigjack *x)
{
	int i, n;
	t_atom form[3];
	form[0].a_type = A_LONG;
	form[1].a_type = form[2].a_type = A_FLOAT;

	for (i = 0; i < x->x_npointsout; i++)
	{
		int valid = (x->x_peaks[i].p_status == S_ON);
		form[0].a_w.w_long = i;
		form[1].a_w.w_float = (valid ? x->x_peaks[i].p_amp : 0.);
		form[2].a_w.w_float = x->x_peaks[i].p_freq;
		outlet_list(x->x_partialoutlet, ps_list, (valid ?3:2), form);
	}

	outlet_float(x->x_poweroutlet, x->x_power);
	outlet_float(x->x_micropitchoutlet, x->x_micropitch);
	if (x->x_attack)
	{
		outlet_bang(x->x_attackoutlet);
		x->x_attack = 0;
	} 
	if (x->x_pitchoutput > 0)
	{
		outlet_float(x->x_pitchoutlet, (double)(x->x_pitchoutput));
		x->x_pitchoutput = 0;
	}
}

void sigjack_clean(t_sigjack *x)
{
	freeobject(x->x_clock);
	t_freebytes(x->x_freeme, x->x_npts * sizeof(float) + ALIGN);
	dsp_free((t_pxobject *)x);
}

void *sigjack_new(long npoints, long period, long npartials, long npointsout)
{
	t_sigjack *x;
	float *fp;
	int logn, i;
	if (!npoints) npoints = 512;
	logn = sigjack_ilog2(npoints);
	if (npoints != (1 << logn))
	{
		error("jack~: power of two required as argument");
		return (0);
	}
	if (npoints > MAXPOINTS || npoints < MINPOINTS)
	{
		error("jack~: argument out of range");
		return (0);
	}
	if (period & (MAXVS-1))
	{
		error("jack~: period must be a multiple of %d", MAXVS);
		return (0);
	}
	if (period > npoints)
	{
		error("jack~: period can't exceed window size");
		return (0);
	}
	if (!period)
		period = npoints;
	if (!(fp = (float *) t_getbytes(npoints * sizeof(float))))
	{
		error("jack~: out of memory");
		return (0);
	}
	x = (t_sigjack *)newobject(sigjack_class);
	
	sigjack_nfilters(x, npartials, npointsout);
	x->x_freeme = fp;
	set_zero(fp, npoints);
	dsp_setup((t_pxobject *)x, 1);

	x->x_partialoutlet = listout(x);
	x->x_poweroutlet = floatout(x);
	x->x_micropitchoutlet = floatout(x);
	x->x_attackoutlet = bangout(x);
	x->x_pitchoutlet = floatout(x);

	x->x_clock = clock_new(x, (method)sigjack_tick);
	//fft_init();
	pt__init();
	
	x->x_srate = 44100.;
	x->x_overlap = npoints-period;
	x->x_buf = x->x_fillbuf = fp;
	x->x_ebuf = x->x_buf + npoints;
	x->x_npts = npoints;
	x->x_nchan = PEROCTAVE * logn - 2 * PEROCTAVE;
	x->x_lofreq = 1.;
	x->x_downsample = 1;
	x->x_printme = 0;
	x->x_reffreq = 440.;
	x->x_maxerror = .4;
	sigjack_stretch(x, 1.);
	sigjack_amp_range(x, OFFPOWTHRESH, ONPOWTHRESH);
	sigjack_quality_range(x, OFFQUALITYTHRESH, ONQUALITYTHRESH);
	sigjack_pitch_range(x, 0, 1000);
	sigjack_reattack(x, DEFREATTACK, DEFRATIME);
	sigjack_debouncetime(x, DEFDEBOUNCETIME);
	x->x_midipit = 0;
	x->x_midilastout = x->x_miditime = 0;
	x->x_lastindex = 0;
	x->x_stick = STICK;
	x->x_peaked = 0;
	x->x_histindex = 0;
	x->x_countup = 1000;
	x->x_attack = 0;
	x->x_peakthresh = .1;
	x->x_micropitch = x->x_pitchoutput = 0;
	for(i = 0; i < HISTORY; i++)
		x->x_pitchhist[i] = x->x_amphist[i] = 0;
	for (i = 0; i < NAMPCOEF; i++)
		x->x_ampcoef[i] = sigjack_defcoef[i];
	sigjack_glisstime(x, 0);
	sigjack_vibrato(x, 0, .75);
	sigjack_loud(x, 0L);
	return (x);
}

#ifdef COHERENCE0
		/* coherence test.	The peak and the two adjacent vectors should
		point in almost opposite directions.	We test whether (det[xy])^2
		is small compared to (|x||y|)^2; this would also be true in the
		(theoretically unlikely) event that x and y were pointing in the
		same direction and not in opposite ones.	We check that the dot
		product is negative to avoid allowing that case. */
	
	threshtimes = hanningpow[1] * x->x_peakthresh;
	det = hanningbuf[0] * hanningbuf[3] - hanningbuf[1] * hanningbuf[2];
 	if (x->x_printme && (nprint > 0))
		post("prev %f %f last %f %f", hanningbuf[0], hanningbuf[1],
			hanningbuf[2], hanningbuf[3]);
	if (det * det > threshtimes * hanningpow[0]) continue;

	det = hanningbuf[4] * hanningbuf[3] - hanningbuf[5] * hanningbuf[2];
 	if (x->x_printme && (nprint > 0))
		post("new %f %f", hanningbuf[4], hanningbuf[5]);
	if (det * det > threshtimes * hanningpow[2]) continue;
#endif


#ifdef NPOINTSOUT0
		int nnewleft = nnewpeaks;	/* number of new peaks after culling */
		t_peak *p1, *p2;
		static int ndidit = 1;

		/* Cull the new peak list, keeping no more than x->x_npointsout+2 peaks.
		If we have more than that, drop the weakest ones. */

		if (ndidit) post("x1 %lx", sigjack_addr());
		 
		for (i = nnewpeaks; i > x->x_npointsout+2; i--)
		{
			float worst;
			for (p1 = 0, p2 = newpeaks, worst = 1E35, j = nnewpeaks; j--; p2++)
			{
	if (p2->p_amp < worst && p2->p_status) p1 = p2, worst = p2->p_amp;
			}
			if (p1)
			{
				p1->p_status = 0;
	nnewleft--;
			}
		}

		/* for each peak in the new spectrum, try to find a peak in the
		"current spectrum" which matches it.	It's probably impossible for
		two peaks to find the same "current" one but if it happens we just
		take either one.
		*/

		if (ndidit) post("x2 %lx", sigjack_addr());

		for (i = 0, p1 = newpeaks; i < nnewpeaks; i++, p1++)
		{
			p1->p_old = 0;
			if (p1->p_status)
			{
	float bin = x->x_srate / x->x_npts;
	t_peak *p2;
	for (p2 = x->x_peaks, j = x->x_npointsout; j--; p2++)
	{
		if (p2->p_status && p2->p_freq < p1->p_freq + bin &&
			p2->p_freq > p1->p_freq - bin)
		{
			p1->p_old = p2;
			break;
		}
	}
			}
		}


			/* initialize current states to zero.	The information
			about continuity is all contained in "newpeaks" as filled
			in above; the new peak points to the matching current peak. */

		if (ndidit) post("x3 %lx", sigjack_addr());
		for (i = 0; i < x->x_maxpeaks; i++)
			x->x_peaks[i].p_status = 0;

	/* mark states of current formants that match a new one as ON.
	Since we just cleared all the current states to zero, the states
	will then be either "ON" or zero. */

		for (i = 0, p1 = newpeaks; i < nnewpeaks; i++, p1++) if (p1->p_old)
		{
			t_peak *p2 = p1->p_old;
			p2->p_status = S_ON;
			p2->p_amp = p1->p_amp;
			p2->p_freq = p1->p_freq;
		}
	

	/* Continue eliminating the weakest peaks, of which no more than
	two should now remain, until nnewleft does not exceed
	x->x_npointsout.	In contrast to the earlier culling session,
	we don't eliminate a peak here if it has been matched. */

		if (ndidit) post("x4 %lx", sigjack_addr());
		while (nnewleft > x->x_npointsout)
		{
			float worst;
			for (p1 = 0, p2 = newpeaks, worst = 1E35, j = nnewpeaks; j--; p2++)
			{
	if (!p2->p_old && p2->p_amp < worst && p2->p_status)
		p1 = p2, worst = p2->p_amp;
			}
			if (!p1) break;	/* this shouldn't happen */
			p1->p_status = 0;
			nnewleft--;
		}
		
			/* now find a home for all the remaining peaks. */

		if (ndidit) post("x5 %lx", sigjack_addr());
			for (p1 = newpeaks, p2 = x->x_peaks, i = nnewpeaks, j = x->x_npointsout;
	i--; p1++)
			{
	if (!p1->p_old && p1->p_status)
	{
		while (j && p2->p_status) j--, p2++;
		if (!j) break;
		p2->p_status = S_NEXTON;
		p2->p_amp = p1->p_amp;
		p2->p_freq = p1->p_freq;
	}
			}
	
		if (ndidit) post("x6 %lx", sigjack_addr());
		if (x->x_printme)
		{
			for (i = 0; i < x->x_npointsout; i++)
	if (x->x_peaks[i].p_status)
		post("%2.2d ---> %7d %7d %d", i,
			(int)x->x_peaks[i].p_freq, 
			(int) (100000 * x->x_peaks[i].p_amp), 
				x->x_peaks[i].p_status);
	else post("%d off", i);
		}
		ndidit	 = 0;
#endif /* 0 */
