#include "ext.h"
#include "z_dsp.h"
#include "buffer.h"

//#define RES_ID 6668

typedef struct _tabpoke
{
    t_pxobject l_obj;
    t_symbol *l_sym;
    t_buffer *l_buf;
    long l_chan;
    float l_outval;
    void *l_outlet;
} t_tabpoke;

long Constrain(long v, long lo, long hi);
void tabpoke_set(t_tabpoke *x, t_symbol *s);
void *tabpoke_new(t_symbol *s, long chan);
void tabpoke_bang(t_tabpoke *x);
void tabpoke_in1(t_tabpoke *x, long n);
void tabpoke_int(t_tabpoke *x, long n);
void tabpoke_list(t_tabpoke *x, Symbol *s, int ac, Atom *av);
void tabpoke_assist(t_tabpoke *x, void *b, long m, long a, char *s);
void tabpoke_dblclick(t_tabpoke *x);

t_symbol *ps_buffer;

void *tabpoke_class;

void main(void)
{
	setup((t_messlist **)&tabpoke_class, (method)tabpoke_new, /*(method)dsp_free*/ 0L, (short)sizeof(t_tabpoke), 0L, 
		A_SYM, A_DEFLONG, 0);
	addbang((method)tabpoke_bang);
	addint((method)tabpoke_int);
	//addinx((method)tabpoke_in1,1); //no channel changing for now
	addmess((method)tabpoke_list, "list", A_GIMME, 0);
	addmess((method)tabpoke_set, "set", A_SYM, 0);
	addmess((method)tabpoke_assist, "assist", A_CANT, 0);
	addmess((method)tabpoke_dblclick, "dblclick", A_CANT, 0);
	//dsp_initclass();
	ps_buffer = gensym("buffer~");
	//rescopy('STR#',RES_ID);
}

long Constrain(long v, long lo, long hi)
{
	if (v < lo)
		return lo;
	else if (v > hi)
		return hi;
	else
		return v;
}


void tabpoke_set(t_tabpoke *x, t_symbol *s)
{
	t_buffer *b;
	
	x->l_sym = s;
	if ((b = (t_buffer *)(s->s_thing)) && ob_sym(b) == ps_buffer) {
		x->l_buf = b;
	} else {
		error("tabpoke~: no buffer~ %s", s->s_name);
		x->l_buf = 0;
	}
}

void tabpoke_list(t_tabpoke *x, Symbol *s, int ac, Atom *av)
{
	float val;
	float *tab;
	long index,chan,frames,nc;
	
	t_buffer *b;
	
	if ((b = (t_buffer *)(x->l_sym->s_thing)) && ob_sym(b) == ps_buffer) {
		x->l_buf = b;
	} else {
		error("tabpoke~: no buffer~ %s", x->l_sym->s_name);
		x->l_buf = 0;
	}
	if (x->l_buf)
	{
	tab = b->b_samples;
	chan = x->l_chan;
	frames = b->b_frames;
	nc = b->b_nchans;
	
	if(ac <=2) {
		if (av[0].a_type == A_LONG)
			index = av[0].a_w.w_long;
		else if (av[0].a_type == A_FLOAT)
			index = (long)(av[0].a_w.w_float+0.5);
		if (av[1].a_type == A_LONG)
			val = (float)av[1].a_w.w_long;
		else if (av[1].a_type == A_FLOAT)
			val = av[1].a_w.w_float;
	}
	if (index < 0)
		index = 0;
	else if (index >= frames)
		index = frames - 1;
	if (nc > 1)
		index = index * nc + chan;
	tab[index] = val;
	}
}

void tabpoke_bang(t_tabpoke *x)
{
	outlet_float(x->l_outlet, (float)x->l_outval);
}

void tabpoke_int(t_tabpoke *x, long n)
{
	float val;
	float *tab;
	long chan,frames,nc;
	long index = n;
	t_buffer *b;
	
	if ((b = (t_buffer *)(x->l_sym->s_thing)) && ob_sym(b) == ps_buffer) {
		x->l_buf = b;
	} else {
		error("tabpoke~: no buffer~ %s", x->l_sym->s_name);
		x->l_buf = 0;
	}
	if (x->l_buf)
	{
	tab = b->b_samples;
	chan = x->l_chan;
	frames = b->b_frames;
	nc = b->b_nchans;
	
	if (index < 0)
		index = 0;
	else if (index >= frames)
		index = frames - 1;
	if (nc > 1)
		index = index * nc + chan;
	x->l_outval = tab[index];
	outlet_float(x->l_outlet, (float)x->l_outval);
	}
}

void tabpoke_in1(t_tabpoke *x, long n)
{
	if (n)
		x->l_chan = Constrain(n,1,4) - 1;
	else
		x->l_chan = 0;
}


void tabpoke_dblclick(t_tabpoke *x)
{
	t_buffer *b;
	
	if ((b = (t_buffer *)(x->l_sym->s_thing)) && ob_sym(b) == ps_buffer)
		mess0((Object *)b,gensym("dblclick"));
}

void tabpoke_assist(t_tabpoke *x, void *b, long m, long a, char *s)
{
	//assist_string(RES_ID,m,a,1,3,s);
	if (m==1) {
		switch (a) {
			case 0: sprintf(s,"list to store, int to recall sample value"); break;
		}
	} 
	else if (m==2) {
		switch (a) {
			case 0: sprintf(s,"sample value out"); break;
		}
	}
}

void *tabpoke_new(t_symbol *s, long chan)
{
	t_tabpoke *x = (t_tabpoke *)newobject(tabpoke_class);
	//dsp_setup((t_pxobject *)x, 0);
	//intin((t_object *)x,1);
	x->l_outlet = floatout(x);
	x->l_sym = s;
	x->l_outval=0;
	chan = 1; // only one channel for now
	tabpoke_in1(x,chan);
	return (x);
}

