

#include "ext.h"
#include "z_dsp.h"
#include "math.h"

void *sigthreshold_class;

typedef struct _sigthreshold
{
	t_pxobject x_obj;
	void *x_outlet1;			/* "bang" out when reaches hithresh */
	void *x_outlet2;			/* "bang" out when reaches lothresh */
	long c_wait;			/* ticks to wait before becoming active */
	float c_hithresh;
	float c_lothresh;
	long c_hideadtime;
	long c_lodeadtime;
	void *c_clock;
	double x_hideadmsec;		/* hi dead time in msec */
	double x_lodeadmsec;		/* lo dead time in msec */
	double x_tickcvt;			/* msec to DSP chain tick conversion */
	short c_on;			/* true if waiting for the low threshold */
} t_sigthreshold;


void sigthreshold_assist(t_sigthreshold *x, void *b, long m, long a, char *s);
t_int *sigthreshold_perform(t_int *w);
void sigthreshold_dsp(t_sigthreshold *x, t_signal **sp);
void sigthreshold_set(t_sigthreshold *x, t_symbol *s, short argc, t_atom *argv);
void sigthreshold_iset(t_sigthreshold *x, double d1, double t1, double d2, double t2);
void sigthreshold_int(t_sigthreshold *x, long n);
void sigthreshold_float(t_sigthreshold *x, double f);
void sigthreshold_tick(t_sigthreshold *x);
void sigthreshold_ff(t_sigthreshold *x);
void *sigthreshold_new(t_symbol *s, short argc, t_atom *argv);

void main(void)
{
	setup((t_messlist **)&sigthreshold_class, (method)sigthreshold_new, (method)sigthreshold_ff, 
		(short)sizeof(t_sigthreshold), 0L, A_GIMME, 0);
	addmess((method)sigthreshold_dsp, "dsp", A_CANT, 0);
	addfloat((method)sigthreshold_float);
	addint((method)sigthreshold_int);
	addmess((method)sigthreshold_assist, "assist", A_CANT, 0);
	addmess((method)sigthreshold_set, "set", A_GIMME, 0);
	dsp_initclass();
}

void sigthreshold_assist(t_sigthreshold *x, void *b, long m, long a, char *s)
{
	if (m==1) {
		switch (a) {
			case 0: sprintf(s,"(Signal) Input to Detect"); break;
		}
	} 
	else if (m==2) {
		switch (a) {
			case 0: sprintf(s,"bang When Threshold Detected"); break;
		}
	}
}

t_int *sigthreshold_perform(t_int *w)
{
	float *in1 = (float *)(w[2]);
	t_sigthreshold *x = (t_sigthreshold *)(w[3]);
    if (*(long *)(w[1]))
	    goto out;

	if (x->c_wait)
		x->c_wait--;
	else if (x->c_on) {
		if (*in1 < x->c_lothresh) {	/* if sample < lo threshold ... */
	
			/* set the clock to go off as soon as possible --
				 i.e., at the next control period */
			clock_delay(x->c_clock, 0L);
			/* set dead time and go into "off" state */
			x->c_on = 0;
			x->c_wait = x->c_lodeadtime;
		}
	} else {
		if (*in1 >= x->c_hithresh) {		/* if sample >= hi threshold ... */
			/* set the clock */
			clock_delay(x->c_clock, 0L);
			/* set dead time and go into "on" state */
			x->c_on = 1;
			x->c_wait = x->c_hideadtime;
		}
	}
out:
	return w + 4;
}

/* routine which FTS calls to put you on the DSP chain or take you off.
	 This is never called explicitly; see the setup routine's use of A_CANT */

void sigthreshold_dsp(t_sigthreshold *x, t_signal **sp)
{
	x->x_tickcvt = sp[0]->s_sr / (1000. * sp[0]->s_n);
	sigthreshold_iset(x, x->c_hithresh, x->x_hideadmsec,
			 x->c_lothresh, x->x_lodeadmsec);
	dsp_add(sigthreshold_perform, 3, &x->x_obj.z_disabled, sp[0]->s_vec, x);
}

void sigthreshold_set(t_sigthreshold *x, t_symbol *s, short argc, t_atom *argv)
{
	double d1, d2, t1, t2;
	
	d1 = argc >= 1 && argv->a_type == A_FLOAT? argv->a_w.w_float : 0.0;
	t1 = argc >= 2 && (argv+1)->a_type == A_FLOAT? (argv+1)->a_w.w_float : 0.0;
	d2 = argc >= 3 && (argv+2)->a_type == A_FLOAT? (argv+2)->a_w.w_float : 0.0;
	t2 = argc >= 4 && (argv+3)->a_type == A_FLOAT? (argv+3)->a_w.w_float : 0.0;
	sigthreshold_iset(x,d1,t1,d2,t2);
}
	
void sigthreshold_iset(t_sigthreshold *x, double d1, double t1, double d2, double t2)
{
	long ticks;
	x->c_hithresh = d1;
	x->x_hideadmsec = t1;
	ticks = (long)(x->x_tickcvt * t1) - 1;
	x->c_hideadtime = (ticks > 0 ? ticks : 0);
	if (d2 > d1)
		d2 = d1;
	x->c_lothresh = d2;
	x->x_lodeadmsec = t2;
	ticks = (long)(x->x_tickcvt * t2) - 1;
	x->c_lodeadtime = (ticks > 0 ? ticks : 0);
}

void sigthreshold_int(t_sigthreshold *x, long n)		/* set state */
{
	x->c_on = n;
	x->c_wait = 0;
}

void sigthreshold_float(t_sigthreshold *x, double f)
{
	sigthreshold_int(x, (long)f);
}

/*
	void sigthreshold_bang(x, f)
	t_sigthreshold *x;
	double f;
	{
	post("hi thresh %f, dead %d",
	x->x_ctl->c_hithresh, x->x_ctl->c_hideadtime);
	post("lo thresh %f, dead %d",
	x->x_ctl->c_lothresh, x->x_ctl->c_lodeadtime);
	post("on %d, wait %d", x->x_ctl->c_on, x->x_ctl->c_wait);
	}
	*/

void sigthreshold_tick(t_sigthreshold *x)		/* callback function for the clock */
{
	if (x->c_on)
		outlet_bang(x->x_outlet1);
	else
		outlet_bang(x->x_outlet2);
}

void sigthreshold_ff(t_sigthreshold *x)		/* cleanup on free */
{
	freeobject(x->c_clock);
	dsp_free((t_pxobject *)x);
}

/* hi thresh, hi deadtime, lo thresh, lo deadtime */
void *sigthreshold_new(t_symbol *s, short argc, t_atom *argv)
{
	t_sigthreshold *x = (t_sigthreshold *)newobject(sigthreshold_class);
	dsp_setup((t_pxobject *)x,1);
	//u_setup((t_ugen *)x, IN1, OUT0);
	//u_setassist((t_ugen *)x, 3225, 1);
	x->c_on = 0;		/* state 0 (low) at startup */
	x->c_wait = 0;		/* no dead time */
	x->c_clock = clock_new(x, (method)sigthreshold_tick);
	x->x_outlet2 = bangout((t_object *)x);
	x->x_outlet1 = bangout((t_object *)x);
	x->x_tickcvt = 0.;		/* fake; it's reset in "put" routine */
	sigthreshold_set(x, s, argc, argv);
	//if (u_debug)
	//	post("sigthreshold %lx", x);
	return (x);
}

