#include "ext.h"
#include "z_dsp.h"

typedef struct _rufus
{
	t_pxobject x_obj;
    long x_setcount;
    long x_actcount;
    long x_powerlock;
    long x_waiting;
    float x_prev;
    double x_outcount;
} t_rufus;

void *rufus_class;

t_int *rufus_perform(t_int *w);
void rufus_dsp(t_rufus *x, t_signal **sp);
void *rufus_new(long setcount);
void rufus_assist(t_rufus *x, void *b, long m, long a, char *s);
void rufus_bang(t_rufus *x);
void rufus_stop(t_rufus *x);
void rufus_int(t_rufus *x, long n);
void rufus_in1(t_rufus *x, long n);
void rufus_list(t_rufus *x, t_symbol *s, short argc, t_atom *argv);
void rufus_float(t_rufus *x, double f);
void rufus_powerlock(t_rufus *x);

void main(void)
{
    setup((t_messlist **)&rufus_class,(method)rufus_new, (method)dsp_free, (short)sizeof(t_rufus), 
    	0L, A_DEFLONG, 0);
    addmess((method)rufus_dsp, "dsp", A_CANT, 0);
    addmess((method)rufus_assist, "assist", A_CANT, 0);
    addbang((method)rufus_bang);
    addmess((method)rufus_powerlock, "powerlock", 0);
    addmess((method)rufus_stop, "stop", 0);
    addint((method)rufus_int);
    dsp_initclass();
}

t_int *rufus_perform(t_int *w)
{
	t_rufus *x = (t_rufus *)(w[1]);
	t_float *in = (t_float *)(w[2]);
	t_float *out = (t_float *)(w[3]);
    int n = (int)(w[4]);
	float prev,val;
	double outcount;
	Boolean starting = false;
	
	if (x->x_obj.z_disabled)
		goto out;
	prev = x->x_prev;
	outcount = x->x_outcount;
	if (x->x_powerlock) {
		while (n--)
			*out++ = *in++;
		goto out;
	} else if (x->x_waiting) {
		while (n--) {
			val = *in++;
			if (prev > val) {
				starting = true;
				x->x_waiting = false;
				prev = val;
				outcount = 0.;
				n++;
				break;
			}
			prev = val;
			*out++ = -1.;
		}
	} else if (x->x_actcount > 0)
		starting = true;
	
	if (starting && n > 0) {
		while (n--) {
			val = *in++;
			if (prev > val)	{ // a sawtooth reset
				if (!--x->x_actcount) {
					prev = val;
					*out++ = -.1;
					break;
				}
			}
			*out++ = outcount++;
			prev = val;
		}
	}
	
	// anything else is over
	
	if (n > 0) {
		while (n--) {
			*out++ = -1.;
		}
	}
	x->x_prev = prev;
	x->x_outcount = outcount;
out:
    return (w+5);
}

void rufus_dsp(t_rufus *x, t_signal **sp)
{
	x->x_prev = 0.;
	x->x_outcount = 0.;
	dsp_add(rufus_perform, 4, x, sp[0]->s_vec, sp[1]->s_vec, sp[0]->s_n);
}

void rufus_assist(t_rufus *x, void *b, long m, long a, char *s)
{
	if (m==1) {
		switch (a) {
			case 0: sprintf(s,"(signal) Input from fft~, int Starts Repeating"); break;
		}
	} 
	else if (m==2) {
		switch (a) {
			case 0: sprintf(s,"(signal) Repeated Input or -1"); break;
		}
	}
}

void rufus_bang(t_rufus *x)
{
	x->x_actcount = x->x_setcount;
	x->x_waiting = true;
	x->x_powerlock = false;
}

void rufus_stop(t_rufus *x)
{
	x->x_actcount = 0;
	x->x_waiting = false;
	x->x_powerlock = false;
}

void rufus_int(t_rufus *x, long n)
{
	if (n > 0) {
		x->x_actcount = x->x_setcount = n;
		x->x_waiting = true;
		x->x_powerlock = false;
	} else
		rufus_stop(x);
}

void rufus_float(t_rufus *x, double f)
{
	rufus_int(x,(long)f);
}

void rufus_powerlock(t_rufus *x)
{
	x->x_powerlock = true;
}

void *rufus_new(long setcount)
{
	t_rufus *x = (t_rufus *)newobject(rufus_class);
	dsp_setup((t_pxobject *)x,1);
	outlet_new((t_object *)x, "signal");
	x->x_setcount = setcount;
	x->x_actcount = 0;
	x->x_waiting = false;
	x->x_powerlock = false;
	return (x);
}
