#include "ext.h"
#include "z_dsp.h"

typedef struct sigsamphold
{
	t_pxobject x_obj;
	float c_val;
	float c_last;
} t_sigsamphold;

void sigsamphold_assist(t_sigsamphold *x, void *b, long m, long a, char *s);
t_int *sigsamphold_perform(t_int *w);
void sigsamphold_dsp(t_sigsamphold *x, t_signal **sp);
void sigsamphold_float(t_sigsamphold *x, double f);
void sigsamphold_int(t_sigsamphold *x, long n);
void sigsamphold_reset(t_sigsamphold *x);
void *sigsamphold_new(void);

void *sigsamphold_class;

void main(void)
{
	setup((t_messlist **)&sigsamphold_class, (method)sigsamphold_new, (method)dsp_free, (short)sizeof(t_sigsamphold), 
		0L, 0);
	addmess((method)sigsamphold_dsp, "dsp", A_CANT, 0);
	addfloat((method)sigsamphold_float);
	addint((method)sigsamphold_int);
	addmess((method)sigsamphold_reset, "reset", 0);
	dsp_initclass();
}

void sigsamphold_assist(t_sigsamphold *x, void *b, long m, long a, char *s)
{
	if (m==1) {
		switch (a) {
			case 0: sprintf(s,"(signal) Control"); break;
			case 1: sprintf(s,"(signal) Value"); break;
		}
	} 
	else if (m==2) {
		switch (a) {
			case 0: sprintf(s,"(signal) Sample/Hold Output"); break;
		}
	}
}

t_int *sigsamphold_perform(t_int *w)
{
	t_sigsamphold *x = (t_sigsamphold *)(w[1]);
	float *in1 = (float *)(w[2]);
	float *in2 = (float *)(w[3]);
	float *out = (float *)(w[4]);
	long n = w[5];
	float last,val,next;
	
	if (x->x_obj.z_disabled)
		goto out;
	
	last = x->c_last;
	val = x->c_val;
	while (n--) {
		next = *(in1++);
		if (next < last) val = *in2;
		in2++;
		last = next;
		*out++ = val;
	}
	x->c_last = last;
	x->c_val = val;
out:
	return w + 6;
}

#ifdef FTS
void sigsamphold_put(t_sigsamphold *x, long whether)
{
	if (whether) {
		u_stdout((t_ugen *)x);
		x->x_ctl = dspchain_addstate(&x->x_cspace,  sizeof(t_sampholdctl), sizeof(float));
		dspchain_addc((d_method)cu_samphold, 5,
			(long)x->x_io[0]->s_shit, (long)x->x_io[1]->s_shit, (long)x->x_io[2]->s_shit,
			(long)x->x_ctl, x->x_io[0]->s_n);
		if (u_debug)
			post("sigsamphold %lx: (%lx,%lx)->%lx",
				x, x->x_io[0]->s_shit, x->x_io[1]->s_shit, x->x_io[2]->s_shit);
	} else {
		bcopy(x->x_ctl, &x->x_cspace, sizeof(t_sampholdctl));
		x->x_ctl = &x->x_cspace;
	}
}
#else
void sigsamphold_dsp(t_sigsamphold *x, t_signal **sp)
{
	dsp_add(sigsamphold_perform,5,x,sp[0]->s_vec,sp[1]->s_vec,sp[2]->s_vec,sp[0]->s_n);
}
#endif

void sigsamphold_float(t_sigsamphold *x, double f)
{
	x->c_val = f;
}

void sigsamphold_int(t_sigsamphold *x, long n)
{
	x->c_val = n;
}

void sigsamphold_reset(t_sigsamphold *x)
{
	x->c_last = 1.e30;
}

void *sigsamphold_new(void)
{
	t_sigsamphold *x = (t_sigsamphold *)newobject(sigsamphold_class);
	dsp_setup((t_pxobject *)x,2);
	outlet_new((t_object *)x,"signal");
	x->c_last = 1.e30;
	x->c_val = 0;
	return (x);
}
